r"""Qubit tapering - operators and Ansatzae."""


# imports
from inquanto.ansatzes import FermionSpaceAnsatzUCCSD
from inquanto.express import load_h5
from inquanto.mappings import QubitMappingJordanWigner
from inquanto.spaces import FermionSpace, QubitSpace
from inquanto.states import FermionState
from inquanto.symmetry import PointGroup, TapererZ2

# Load the fermionic Hamiltonian, define HF state and map to qubits
h2 = load_h5("h2_sto3g.h5", as_tuple=True)
hamiltonian_operator = h2.hamiltonian_operator.to_FermionOperator()
fermion_space = FermionSpace(4)
hf_state = FermionState([1, 1, 0, 0])
qubit_space = QubitSpace(4)
qubit_hamiltonian = QubitMappingJordanWigner.operator_map(hamiltonian_operator)
qubit_hf_state = QubitMappingJordanWigner.state_map(hf_state)

# Generate the symmetry operators from the Hamiltonian and find the symmetry sector of the HF state
symmetry_operators = qubit_space.symmetry_operators_z2(qubit_hamiltonian)
symmetry_sector = [x.symmetry_sector(qubit_hf_state) for x in symmetry_operators]

# We generate a TapererZ2 object, providing it information regarding the symmetry operator and symmetry sector.
taperer = TapererZ2(symmetry_operators, symmetry_sector)

# Now we can taper the Hamiltonian - note that three qubits have been tapered off.
tapered_hamiltonian = taperer.tapered_operator(qubit_hamiltonian)
print("Original Hamiltonian: {}".format(qubit_hamiltonian))
print("Tapered Hamiltonian: {}".format(tapered_hamiltonian))

# ###### OUTPUT #######
# Original Hamiltonian: (-0.059620582760346125) [] + (0.1757594291831965) [Z0] + (-0.23667117678035662) [Z2] + (
# 0.1757594291831965) [Z1] + (-0.23667117678035662) [Z3] + (0.12222714936261847) [Z0 Z2] + (0.12222714936261847) [Z1 Z3] + (0.170015464396032) [Z0
# Z1] + (0.1671443192533724) [Z1 Z2] + (0.044917169890753915) [Y0 X1 X2 Y3] + (-0.044917169890753915) [X0 X1 Y2 Y3] + (-0.044917169890753915) [Y0
# Y1 X2 X3] + (0.044917169890753915) [X0 Y1 Y2 X3] + (0.1671443192533724) [Z0 Z3] + (0.17570338331907048) [Z2 Z3]
# Tapered Hamiltonian: (-0.292644672277225) [I2] + (-0.824861211927106) [Z2] + (0.179668679563016) [X2]

# We can also optionally relabel the qubits in our tapered Hamiltonian
tapered_hamiltonian = taperer.tapered_operator(qubit_hamiltonian, relabel_qubits=True)
print("Tapered Hamiltonian: {}".format(tapered_hamiltonian))

# Output: Tapered Hamiltonian: (-0.292644672277225) [I0] + (-0.824861211927106) [Z0] + (0.179668679563016) [X0]

# We can additionally generate a taperered Ansatz.  The behaviour of how to deal with symmetry-breaking excitations can be controlled.  By default,
# each excitation will be checked, and an exception will be thrown if a symmetry-violating excitation is found:

ansatz_untapered = FermionSpaceAnsatzUCCSD(fermion_space, hf_state)

try:
    _ = FermionSpaceAnsatzUCCSD(fermion_space, hf_state, taperer=taperer)
except ValueError:
    print("This doesn't work because our Ansatz has symmetry breaking excitations!")

# However, we can instead discard them:
# InQuanto will warn the user
ansatz_tapered_discarding = FermionSpaceAnsatzUCCSD(
    fermion_space,
    hf_state,
    taperer=taperer,
    tapering_exponent_check_behaviour="discard",
)

print("#### UNTAPERED ANSATZ REPORT ####")
print(ansatz_untapered.generate_report())
print("\n#### TAPERED ANSATZ REPORT, SYMMETRY VIOLATING EXCITATIONS DISCARDED ####")
print(ansatz_tapered_discarding.generate_report())

######## OUTPUT ########
#### UNTAPERED ANSATZ REPORT ####
# {'ansatz_circuit_depth': 58, 'ansatz_circuit_gates': 110, 'ansatz_parameters': 3, 'ansatz_qubits': 4}
#### TAPERED ANSATZ REPORT, SYMMETRY VIOLATING EXCITATIONS DISCARDED ####
# {'ansatz_circuit_depth': 7, 'ansatz_circuit_gates': 7, 'ansatz_parameters': 1, 'ansatz_qubits': 1}
#########################

# Alternatively, if we are confident that we are not generating any symmetry-violating excitations (for instance, by using the symmetry filtering
# of FermionSpaceAnsatzUCCSD), we can skip the (relatively slow) validation entirely:

symmetrised_space = fermion_space = FermionSpace(
    4, point_group=PointGroup("D2h"), orb_irreps=["Ag", "Ag", "B1u", "B1u"]
)

ansatz_tapered_skipping = FermionSpaceAnsatzUCCSD(
    fermion_space,
    hf_state,
    taperer=taperer,
    tapering_exponent_check_behaviour="skip",
)

print("\n#### TAPERED ANSATZ REPORT, SYMMETRY VALIDATION SKIPPED ####")
print(ansatz_tapered_skipping.generate_report())

######## OUTPUT ########
#### TAPERED ANSATZ REPORT, SYMMETRY VALIDATION SKIPPED ####
# {'ansatz_circuit_depth': 7, 'ansatz_circuit_gates': 7, 'ansatz_parameters': 1, 'ansatz_qubits': 1}
#########################
